/*
  rfc2445-test.cc

  Copyright (C) 2011 Taiji Yamada <taiji@aihara.co.jp>
 */
#include <iostream>
#include <vector>
#include "date_time_numeric.hpp"
#include <boost/format.hpp>
#include <boost/algorithm/string.hpp>	// split
#if defined(USE_BOOST_XPRESSIVE)
#include "boost_xpressive_utility.hpp"
using namespace boost::xpressive;
typedef string_sregex_map string_regex_map;
#else
#include "boost_regex_utility.hpp"
#endif
#include <cstdio>
#include <cstdlib>
#include <cstring>
using namespace std;
using namespace boost;

int main()
{
  FILE *ifp;
  char buf[1024];
  {

    if (!(ifp=popen("[ -f rfc2445.txt ] || { curl -RO -C - 'http://www.ietf.org/rfc/rfc2445.txt'; } && cat rfc2445.txt | sed -ne '/^[0-9.]* Recurrence Component Properties$/,/^[0-9.]* Alarm Component Properties$/p'", "r"))) {
      exit(1);
    }
  }

  const char *vcalendar_fmt_ol2k[] = {
    "BEGIN:VCALENDAR\r\n\
PRODID:-//Taiji Yamada <taiji@aihara.co.jp>//tsv2ics//EN\r\n\
METHOD:PUBLISH\r\n\
VERSION:2.0\r\n\
BEGIN:VTIMEZONE\r\n\
TZID:Japan\r\n\
BEGIN:STANDARD\r\n\
DTSTART:19390101T000000\r\n\
TZOFFSETFROM:+0900\r\n\
TZOFFSETTO:+0900\r\n\
TZNAME:JST\r\n\
END:STANDARD\r\n\
END:VTIMEZONE\r\n\
",
    "END:VCALENDAR\r\n"
  }, *vevent_fmt_ol2k[] = {
    "BEGIN:VEVENT\r\n\
DTSTAMP:%1%\r\n\
SUMMARY%2%\r\n\
LOCATION%3%\r\n\
DESCRIPTION%4%\r\n\
DTSTART%5%\r\n\
DTEND%6%\r\n\
CATEGORIES%7%\r\n\
TRANSP%8%\r\n\
PRIORITY:0\r\n\
CLASS:PUBLIC\r\n",
    "END:VEVENT\r\n",
    ";TZID=Japan:%1%T000000",
    ";TZID=Japan:%1%"
  }, *vcalendar_fmt_ical[] = {
    "BEGIN:VCALENDAR\r\n\
PRODID:-//Taiji Yamada <taiji@aihara.co.jp>//tsv2ics//EN\r\n\
METHOD:PUBLISH\r\n\
VERSION:2.0\r\n\
X-WR-TIMEZONE:Asia/Tokyo\r\n\
CALSCALE:GREGORIAN\r\n\
",
    "END:VCALENDAR\r\n"
  }, *vevent_fmt_ical[] = {
    "BEGIN:VEVENT\r\n\
DTSTAMP:%1%\r\n\
SUMMARY%2%\r\n\
LOCATION%3%\r\n\
DESCRIPTION%4%\r\n\
DTSTART%5%\r\n\
DTEND%6%\r\n\
CATEGORIES%7%\r\n\
TRANSP%8%\r\n\
PRIORITY:0\r\n\
CLASS:PUBLIC\r\n",
    "END:VEVENT\r\n",
    ";VALUE=DATE:%1%",
    ";TZID=Asia/Tokyo:%1%"
  };
  const char **vcalendar_fmt = vcalendar_fmt_ical, **vevent_fmt = vevent_fmt_ical;
  struct {
    string summary, location, description, categories, transp;
    posix_time::ptime dtstart, dtend;
    string rdate, rrule/*, exrule, related_to*/;
    vector<string> exdate;
  } vevent = { "", ":My Room", "", ":rrule-test", ":OPAQUE", }, vevent0 = vevent;
  int vevent_count = 0;

  string_regex_map regex_map;
  string sb, line, line_buffer;
  while (fgets(buf, sizeof(buf), ifp)) {
    sb = buf;
    smatch m;
    if (regex_match(sb, m, regex_map["^\\s+==>.*$"])) {
      line_buffer = line;
      line = sb;
    }
    else if (regex_match(sb, m, regex_map["^\\s+(DTSTART[^:]*|DTEND[^:]*|RDATE[^:]*|RRULE"/*|EXRULE|RELATED-TO*/"|EXDATE[^:]*):.*$"])) {
      line += sb;
    }
    else if (regex_match(sb, m, regex_map["^\\s+$"])) {
      line_buffer = line;
      line = "";
    }
    else if (regex_match(sb, m, regex_map["^\\s+or\\s*$"])) {
      line_buffer = line;
      line = "";
    }
    else if (line != "") {
      line += sb;
    }
    if (line_buffer == "") continue;
    string line;
    if (regex_match(line_buffer, m, regex_map["^\\s+(DTSTART[^:]*|DTEND[^:]*|RDATE[^:]*|RRULE"/*|EXRULE|RELATED-TO*/"|EXDATE[^:]*):.*$"])) {
      vector<string> lines;
      split(lines, line_buffer, is_any_of("\n"));
      for (vector<string>::iterator it=lines.begin(); it!=lines.end(); ++it)
	line += regex_replace(*it, regex_map["^\\s{4,5}"], "") + "\r\n";
      line = regex_replace(line, regex_map["\\r\\n "], "");

      if (!vevent_count) cout << format(vcalendar_fmt[0]);

      split(lines, line, is_any_of("\r\n"));
      for (vector<string>::iterator it=lines.begin(); it!=lines.end(); ++it) {
	if (*it == "") continue;
	if (regex_match(*it, m, regex_map["^SUMMARY(.*)$"])) {
	  vevent.summary = m[1];
	}
	else if (regex_match(*it, m, regex_map["^DTSTART;VALUE=DATE:(.*)$"])) {
	  vevent.dtstart = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(24, 0, 0);
	}
	else if (regex_match(*it, m, regex_map["^DTSTART[^:]*:(.*)$"])) {
	  vevent.dtstart = posix_time::from_iso_string(m[1]);
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(24, 0, 0);
	}
	else if (regex_match(*it, m, regex_map["^DTEND;VALUE=DATE:(.*)$"])) {
	  vevent.dtend = posix_time::ptime(gregorian::from_undelimited_string(m[1]));
	}
	else if (regex_match(*it, m, regex_map["^DTEND[^:]*:(.*)$"])) {
	  vevent.dtend = posix_time::from_iso_string(m[1]);
	}
	else if (regex_match(*it, m, regex_map["^RRULE(.*)$"])) {
	  vevent.rrule = m[1];
	}
	/*
	else if (regex_match(*it, m, regex_map["^EXRULE(.*)$"])) {
	  vevent.exrule = m[1];
	}
	else if (regex_match(*it, m, regex_map["^RELATED-TO(.*)$"])) {
	  vevent.related_to = m[1];
	}
	*/
	else if (regex_match(*it, m, regex_map["^RDATE(.*)$"])) {
	  vevent.rdate = m[1];
	}
	else if (regex_match(*it, m, regex_map["^EXDATE(.*)$"])) {
	  vevent.exdate.push_back(m[1]);
	}
      }
      {
	/*
	  iCal results
	  --------
	  RDATE			- entry ignored
	  EXDATE:date,date,..	- entry ignored
	  RRULE: BYHOUR=	- property ignored
	  RRULE: BYMINUTE=	- property ignored
	 */
	vevent.summary = ":Schedule" + lexical_cast<string>(vevent_count);
	if (!vevent.exdate.empty() && vevent.dtstart == posix_time::ptime() && vevent.rrule.empty()) {	// (EXDATE):19960402T010000Z,19960403T010000Z,19960404T010000Z
	  vevent.dtstart = posix_time::from_iso_string("19960401T010000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(2, 0, 0);
	  vevent.rrule = ":FREQ=DAILY;COUNT=10";
	}
	else if (vevent.rdate == ":19970714T123000Z" && vevent.dtstart == posix_time::ptime()) {
	  vevent.dtstart = posix_time::from_iso_string("19970707T123000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(0, 30, 0);
	}
	else if (vevent.rdate == ";TZID=US-EASTERN:19970714T083000" && vevent.dtstart == posix_time::ptime()) {
	  vevent.dtstart = posix_time::from_iso_string("19970630T083000");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(0, 30, 0);
	}
	else if (vevent.rdate == ";VALUE=DATE:19970101,19970120,19970217,19970421,19970526,19970704,19970901,19971014,19971128,19971129,19971225" && vevent.dtstart == posix_time::ptime()) {	// (RDATE);VALUE=DATE:19970101,19970120,19970217,19970421,19970526,19970704,19970901,19971014,19971128,19971129,19971225
	  vevent.dtstart = posix_time::from_iso_string("19970101T000000");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(3, 0, 0);
	}
	else if (!vevent.rdate.empty() && vevent.dtstart == posix_time::ptime()) {	// (RDATE);VALUE=PERIOD:19960403T020000Z/19960403T040000Z,19960404T010000Z/PT3H
	  vevent.dtstart = posix_time::from_iso_string("19960402T000000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(1, 0, 0);
	}
	else if (vevent.rrule == ":FREQ=DAILY;UNTIL=20000131T090000Z;BYMONTH=1" && vevent.dtstart == posix_time::ptime()) {
	  vevent.dtstart = posix_time::from_iso_string("19990101T090000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(4, 0, 0);
	}
	else if (vevent.rrule == ":FREQ=WEEKLY;COUNT=10;WKST=SU;BYDAY=TU,TH" && vevent.dtstart == posix_time::ptime()) {
	  vevent.dtstart = posix_time::from_iso_string("20000303T000000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(24, 0, 0);
	}
	else if (!vevent.rrule.empty() && vevent.dtstart == posix_time::ptime()) {	// (RRULE):FREQ=MINUTELY;INTERVAL=20;BYHOUR=9,10,11,12,13,14,15,16
	  vevent.dtstart = posix_time::from_iso_string("20000404T090000Z");
	  vevent.dtend = vevent.dtstart + posix_time::time_duration(0, 5, 0);
	}
	{
	  ostringstream os;
	  os << format(vevent_fmt[0]) % "20111213T1200Z" %
	    (!vevent.summary.empty() ? vevent.summary : string(":")) %
	    (!vevent.location.empty() ? vevent.location : string(":")) %
	    (!vevent.description.empty() ? vevent.description : string(":")) %
	    ((vevent.dtstart.time_of_day() == posix_time::time_duration()) ? format(vevent_fmt[2]) % gregorian::to_iso_string(vevent.dtstart.date()) : format(vevent_fmt[3]) % posix_time::to_iso_string(vevent.dtstart)) %
	    ((vevent.dtend.time_of_day() == posix_time::time_duration()) ? format(vevent_fmt[2]) % gregorian::to_iso_string(vevent.dtend.date()) : format(vevent_fmt[3]) % posix_time::to_iso_string(vevent.dtend)) %
	    (!vevent.categories.empty() ? vevent.categories : string(":")) %
	    (!vevent.transp.empty() ? vevent.transp : string(":"));
	  if (!vevent.rdate.empty())
	    os << format("RDATE%1%") % vevent.rdate << "\r\n";
	  if (!vevent.exdate.empty())
	    for (vector<string>::iterator it=vevent.exdate.begin(); it!=vevent.exdate.end(); ++it)
	      os << format("EXDATE%1%") % *it << "\r\n";
	  if (!vevent.rrule.empty())
	    os << format("RRULE%1%") % vevent.rrule << "\r\n";
	  string s = regex_replace(os.str(), regex_map["\\\\r\\\\n"], "\r\n");
	  cout << s << format(vevent_fmt[1]);
	}
      }
      ++vevent_count;
      vevent = vevent0;
    }
    //cerr << line_buffer << endl;
    line_buffer = "";
  }
  if (vevent_count) cout << format(vcalendar_fmt[1]);
  return 0;
}
